from setuptools import setup, find_packages

setup(
    name = 'pywapor',
    version = '3.5.16',
    url = 'https://www.fao.org/aquastat/py-wapor/',
    author = "FAO",
    author_email = "bert.coerver@fao.org",
    license = "Apache",
    packages = find_packages(include = ['pywapor', 'pywapor.*']),
    include_package_data=True,
    python_requires='>=3.10',
    install_requires = [
        # 'libgdal-netcdf',
        # 'libgdal-jp2openjpeg',
        'netcdf4',
        'gdal>=3,<4.0',
        'xarray',
        'numpy<2',
        'pandas<3',
        'requests',
        'matplotlib',
        'pyproj',
        'scipy',
        'pycurl',
        'joblib',
        'bs4',
        'rasterio',
        'bottleneck',
        'tqdm',
        'dask',
        'rioxarray>=0.15.5',
        'cryptography',
        'cachetools',
        'cdsapi>=0.7.0',
        'shapely',
        'lxml',
        'scikit-learn',
        'numba',
        'xmltodict',
    ],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)